import pandas as pd
import requests
import logging
import time, sys, os
from functools import wraps
import shutil
import contextlib
from typing import Tuple
import xlrd

# retry decorator with adjustable nbr of retries and retry attempt passed to wrapped function as keyword argument 'attempt'
def retry(retries: int, **fkwargs):
    def _decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for attempt in range(retries):
                try:
                    return func(*args, attempt = attempt, **kwargs)
                except Exception as e:
                    logging.warning(f"Error: {fkwargs['error']}, Attempt: {attempt}, Python Error: {e}")
                    time.sleep(attempt*attempt)
            logging.error("3 attempts to download bpost xls to df failed. Exiting...")
            sys.exit(1)
        return wrapper
    return _decorator

class basicClient():

    def __init__(self) -> None:
        None

    @retry(retries = 3, error = "Error downloading xls file from bpost to df")
    def download_xls_to_df(self, url: str, file_name: str, **kwargs) -> Tuple[str, pd.DataFrame]:
        """ 
            ARGUMENTS
                url: url of the xls file, eg. www.example.com/file.xls
                file_name: nam for storing file, eg. local_file

            RETURNS
                Tuple[local_file, df] where local_file = ./tmp_data/timestamp_file_name.xls and 
                df is the first worksheet read into a dataframe
        
        """
        logging.info(f"Attempt number: {kwargs['attempt']}")
        download_path = "./tmp_data/"
        with contextlib.suppress(Exception):
            shutil.rmtree(download_path)
        timestamp = round(time.time(), 4)
        if not os.path.exists(download_path):
            os.mkdir(download_path)
        temp_location = f"{download_path}{timestamp}_{file_name}.xls"
        response = requests.get(url)
        with open(temp_location, 'wb') as f:
            f.write(response.content)
        workbook = xlrd.open_workbook(temp_location, ignore_workbook_corruption=True)
        df = pd.read_excel(workbook)
        return temp_location, df




from selenium.webdriver.chrome.options import Options
from selenium import webdriver
from selenium.webdriver.common.by import By
import logging
import sys
import os
import time
import urllib

class seleniumClient():

    def __init__(self, env: str, url:str) -> None:
        """
            Initalizes new selenium client. Client allows to download files from webpages. 
            
            env: local, dev or prd. Use local in case of local development in order to correctly create a chromedriver.
            url: url of the page where you want to download a file
        """
        self.env = env
        self.url = url
        self.driver = self._create_driver()
        self.driver.set_window_size(800, 600)

    def _set_chrome_options(self) -> None:
        """
            Sets chrome options for Selenium.
            Chrome options for headless browser is enabled.
            based on https://github.com/nazliander/scrape-nr-of-deaths-istanbul/blob/master/app.py
        """
        # chrome_options = Options()
        download_file_path = os.path.join(os.getcwd(), 'temp_data')
        path_exists = os.path.exists(download_file_path)
        if not path_exists:
            os.makedirs(download_file_path)
        options = webdriver.ChromeOptions()
        options.add_argument("--headless=new")
        options.add_argument("--no-sandbox")
        options.add_argument("--disable-dev-shm-usage")
        prefs = {
            "download.default_directory": download_file_path,
        }

        options.add_experimental_option("prefs",prefs)

        # chrome_prefs = {
        #     "download": {"default_directory": "./temp_data"},
        #     "profile.default_content_settings": {"images": 2},
        # }
        # chrome_options.add_experimental_option("prefs", chrome_prefs)
        return options

    def _create_driver(self):
        logging.info("Creating chrome driver...")
        try: 
            if self.env == 'local':
            # chrome iterates very quickly, hence chrome and chromedriver diverge. Hence, when running local, we can ensure match between both with below code
                from webdriver_manager.chrome import ChromeDriverManager
                driver = webdriver.Chrome(ChromeDriverManager().install(), options=self._set_chrome_options())
            else:
                # driver and chrome are installed in docker image so we can use standard way of creating webdriver
                driver = webdriver.Chrome(options=self._set_chrome_options())
        except Exception as error: 
            logging.error('Issue creating chrome driver. Exiting with status 1')
            logging.error(error)
            sys.exit(1)
        return driver

    def _create_local_dir(self, file_name:str, file_extension: str):
        """
            creates local path ./temp_data/
            creates file within local path: 165384.00_file_name.file_extension 

            returns local_path, file 
        """
        logging.info("Creating local directory to temporarily store downloaded files")
        timestamp = round(time.time(), 4)
        file = f'{timestamp}_{file_name}.{file_extension}'
        local_path = './temp_data/'
        if not os.path.exists(local_path):
                os.mkdir(local_path)
        return f'{local_path}{file}'

    def get_file(self, element_type:str, element_name:str, file_name:str, file_extension:str):
        """
            element_type: 
                ID = "id"
                NAME = "name"
                XPATH = "xpath"
                LINK_TEXT = "link text"
                PARTIAL_LINK_TEXT = "partial link text"
                TAG_NAME = "tag name"
                CLASS_NAME = "class name"
                CSS_SELECTOR = "css selector"
            element_name: 
                eg. <element id="element_name_1'/> has element_name = element_name_1 and element_type=ID
            file_name:
                provide a name for the file that will be downloaded
            file_extension: 
                provide the desired extension

            This function creates a local path ./temp_data/ and creates a file within local path: 165384.00_file_name.file_extension
            Next, the driver downloads the specified file via href by accessing the HTML element specified by element_type and element_name. Retry pattern 
            with exponential back of is implemented. 

            Function returns local_file
        """ 
        for i in range(3):
        # retry pattern with some back off
            time.sleep(i*i)
            try:
                local_file = self._create_local_dir(file_name, file_extension)
                self.driver.get(self.url)
                element = self.driver.find_element(getattr(By, element_type), element_name)
                file_url = element.get_attribute('href')
                daily_file = urllib.request.URLopener()
                daily_file.retrieve(file_url, local_file )
                break
            except Exception as e:
                if i < 2:
                    logging.warning(f'Issue downloading file, trying again: {e}')
                    continue
                elif i == 2: 
                    logging.error('Cannot download file. Exiting..')
                    sys.exit(1)
        logging.info(f'Succesfully retrieved file and stored here: {local_file}')
        return local_file
        
    def get_file_via_form_button_xpath(self, selenium_xpath:str ):
        """
            ARGS: 
                xpath: //form[@id='csvdownload']/button[1]
            
            Donwloads file to temp_data and waits 10 seconds to complete download
        """ 
        for i in range(3):
        # retry pattern with some back off
            time.sleep(i*i)
            try:
                self.driver.get(self.url)
                # safety wait
                time.sleep(1)
                button = self.driver.find_element(By.XPATH, selenium_xpath )
                print(button)
                button.click()
                # REFACTOR - we need to wait before closing the driver, otherwise we end up with an incomplete crcdownload
                # we should check in temp_data whether the filetype is csv
                time.sleep(10)
                break
            except Exception as e:
                if i < 2:
                    logging.warning(f'Issue downloading file, trying again: {e}')
                    continue
                elif i == 2: 
                    logging.error('Cannot download file. Exiting..')
                    sys.exit(1)
        logging.info('Succesfully retrieved file and stored in ./temp_data/')
        return
      
    def quit(self):
        self.driver.quit()
    

    

  



import os, sys
import pandas as pd
import pyodbc
import boto3
import logging
from time import time
from io import BytesIO

class sqlTerraClient():

    DRIVER = '{ODBC Driver 17 for SQL Server}'
    PORT = 1433

    def __init__(self, env):
        """
            When deploying, environment variables are injected depending on ENV via entrypoint.sh.
            For local development, we fetch environemnt variables from local enviroment where we have variables for local, dev and prd, hence the except statement

            ! since we don't have a local redshift cluster we load into redshift dev 
        """
        try:
            logging.info('Trying to fetch sql credentials from environment variables')
            server_raw=os.environ['TERRA_HOST']
            db=os.environ['TERRA_DB']
            username=os.environ['TERRA_USERNAME']
            password=os.environ['TERRA_PASSWORD'] 
        except Exception:
            logging.info('Trying to fetch ENV specific sql credentials from environment variables')
            server_raw=os.environ[f'TERRA_HOST_{env.upper()}']
            db=os.environ[f'TERRA_DB_{env.upper()}']
            username=os.environ[f'TERRA_USERNAME_{env.upper()}']
            password=os.environ[f'TERRA_PASSWORD_{env.upper()}'] 
        server = f'{server_raw},{self.PORT}'
        connection_string = 'DRIVER={ODBC Driver 17 for SQL Server};SERVER=' + server + ';DATABASE=' + db + ';UID=' + username + ';PWD=' + password
        try:
            logging.info('Trying to create a connection to DB')
            self.conn = pyodbc.connect(connection_string)
            logging.info('Succesfully creaeted DB connection')
        except Exception as e:
            logging.error(f'Issue creating connection: {e} .Exiting...')
            sys.exit(1)
        return 

    def query_to_df(self, query:str) -> pd.DataFrame:
        """
            fetchall() where all rows will be stored in memory
            returns pandas dataframe
        """
        logging.info('Executing query on DB')
        cursor = self.conn.cursor()
        cursor.execute(query)
        rows = cursor.fetchall()
        columns = [column[0] for column in cursor.description]
        cursor.close()
        df = pd.DataFrame.from_records(rows, columns=columns)
        logging.info('Succesfully queried DB and returned as a pandas DF')
        return df

    def stream_to_s3_parquet(self, query:str, batch_size:int, s3_bucket:str, s3_prefix:str, s3_filename:str) -> None:
        """
            Streams the results of a sql query to parquet files on s3 with 'batch_size' nbr of rows per file. 
            Output files have the following key:
                {s3_bucket}{s3_prefix}{timestamp}_{filename}.parquet        
        """
        logging.info(f'Streaming result of query {query} to s3 {s3_bucket}/{s3_prefix} with batch size {batch_size}')
        for x in self._stream_results(query, batch_size):
            col_names = [column[0] for column in x[0].cursor_description]
            df = pd.DataFrame.from_records(x, columns=col_names)
            self._df_to_parquet_s3(df, s3_bucket, s3_prefix, s3_filename)
        logging.info('Completed streaming of query to parquet files on s3')
        return
                
    def _stream_results(self, query:str, batch_size: int):
        cursor = self.conn.cursor()
        iterator = cursor.execute(query)
        while True:
            rows = iterator.fetchmany(batch_size)
            if not rows:
                break
            yield rows
            # for row in rows:
            #     yield row

    def _df_to_parquet_s3(self, df:pd.DataFrame, s3_bucket: str, s3_prefix: str, file_name:str):
        parquet_buffer = BytesIO()
        df.to_parquet(parquet_buffer, index=False, allow_truncated_timestamps=True)
        s3 = boto3.resource('s3')
        timestamp = round(time(), 4)
        s3_key = f"{s3_prefix}{timestamp}_{file_name}.parquet"
        s3.Object(s3_bucket, s3_key).put(Body=parquet_buffer.getvalue())
        logging.info(f'Stored {s3_key} on s3 {s3_bucket}')
        return



import pandas as pd
import os
import logging
import sys
import pyodbc
from time import time
from io import BytesIO
import boto3
from time import sleep
from .custom_decorators import retry

class lynxClient():
    """
        Class which creates a connection to SQL server based on connection details in ENV variables.
        Connection and cursor objects are created on initialization with retry patterns. 

        The connection object allows querying and several methods are available which all have retry mechanism built-in. Each method 
        generates a cursor object which is closed after correct execution. Each method is automatically retried 3 times. 

        It is advised to call 'close_connection' if you no longer need the class instance. 
    """

    def __init__(self) -> str:
        self._connection_instance = None
        self._connection_instance = self._connect()
        logging.info('successfully created connection')

    def _create_conn_string(self) -> str:
        """
            Function to create a connection string for SQL server
            Environment variables are injected in the container environment via entrypoint.sh at runtime
            For local development, we fetch environment variables from local enviroment. Make sure they are set up. 
            Lynx only has 1 environment, hence no need to pass ENV.
        """
        try: 
            server_raw = os.environ['LYNX_SERVER']
            port = os.environ['LYNX_PORT']
            database = os.environ['LYNX_DATABASE']
            username = os.environ['LYNX_USERNAME']
            password = os.environ['LYNX_PASSWORD']
            server = f'{server_raw},{port}'
            logging.info("Fetched lynx credentials from environment variables")
        except Exception as e:
            logging.error("Issue fetching lynx credentials from environment variables. Exiting...")
            logging.error(e)
            sys.exit(1)
        connection_string = 'DRIVER={ODBC Driver 17 for SQL Server};SERVER=' + server + ';DATABASE=' + database + ';UID=' + username + ';PWD=' + password
        return connection_string

    @retry(retries=3, error="Error creating connection to SQL server")
    def _connect(self, **kwargs) -> pyodbc.Connection:
        return pyodbc.connect(self._create_conn_string())

    @retry(retries=3, error="Error creating cursor")
    def _create_cursor(self, **kwargs) -> pyodbc.Cursor:
        return self._connection_instance.cursor()

    @retry(retries=3, error="Error executing query to list")
    def query_to_list(self, query:str, **kwargs) :
        """
            fetchall() where all rows will be stored in memory
            returns 
                data list
                columns list
                dtypes list
        """
        cursor = self._create_cursor()
        cursor.execute(query)
        rows = cursor.fetchall()
        # https://stackoverflow.com/questions/5504340/python-mysqldb-connection-close-vs-cursor-close
        columns = [column[0] for column in cursor.description]
        dtypes = [column[1] for column in cursor.description]
        cursor.close()
        return rows, columns, dtypes

    @retry(retries=3, error="Error executing query_fetch_single_value")
    def query_fetch_single_value(self, query: str, **kwargs) -> str:
        """
            grab a scalar, eg select max(x) from y
        """
        cursor = self._create_cursor()
        cursor.execute(query)
        val = cursor.fetchval()
        cursor.close()
        return str(val)

    @retry(retries=3, error="Error executing query_to_df")
    def query_to_df(self, query:str, **kwargs) -> pd.DataFrame:
        """
            fetchall() where all rows will be stored in memory
            returns pandas dataframe
        """
        cursor = self._create_cursor()
        cursor.execute(query)
        rows = cursor.fetchall()
        columns = [column[0] for column in cursor.description]
        cursor.close()
        df = pd.DataFrame.from_records(rows, columns=columns)
        logging.info(df)
        return df

    @retry(retries=3, error="Error executing get_max_cursors ")
    def _get_max_cursors(self, **kwargs) -> str:
        cursor = self._create_cursor()
        nbr_cursors = cursor.getinfo(pyodbc.SQL_MAX_CONCURRENT_ACTIVITIES)
        return nbr_cursors

    @retry(retries=3, error="Error executing stream_to_s3_parquet")
    def stream_to_s3_parquet(self, query:str, batch_size:int, s3_bucket:str, s3_prefix:str, s3_filename:str, **kwargs) -> None:
        """
            Streams the results of a sql query to parquet files on s3 with 'batch_size' nbr of rows per file. 
            Output files have the following key:
                {s3_bucket}{s3_prefix}{timestamp}_{filename}.parquet        
        """
        if kwargs['attempt'] > 1:       # ensure idempotency in case of retry
            s3 = boto3.resource('s3')
            bucket = s3.Bucket(s3_bucket)
            bucket.objects.filter(Prefix=s3_prefix).delete()
        for x in self._stream_results(query, batch_size):
            col_names = [column[0] for column in x[0].cursor_description]
            df = pd.DataFrame.from_records(x, columns=col_names)
            self._df_to_parquet_s3(df, s3_bucket, s3_prefix, s3_filename)
        return
         
    def _stream_results(self, query:str, batch_size: int) -> None:
        cursor = self._create_cursor()
        cursor_iterator = cursor.execute(query)
        while True:
            rows = cursor_iterator.fetchmany(batch_size)
            if not rows:
                break
            yield rows
            # for row in rows:
            #     yield row

    def _df_to_parquet_s3(self, df:pd.DataFrame, s3_bucket: str, s3_prefix: str, file_name:str) -> None:
        parquet_buffer = BytesIO()
        df.to_parquet(parquet_buffer, index=False, allow_truncated_timestamps=True)
        s3 = boto3.resource('s3')
        timestamp = round(time(), 4)
        s3_key = f"{s3_prefix}{timestamp}_{file_name}.parquet"
        s3.Object(s3_bucket, s3_key).put(Body=parquet_buffer.getvalue())
        logging.info(f'Stored {s3_key} on s3 {s3_bucket}')
        del df
        return
    
    def close_connection(self) -> None:
        self._connection_instance.close()
        logging.info("Closed connection to SQL server")
        return



        


import yaml
import os
import sys, inspect
import logging
from attrdict import AttrDict   # https://pypi.org/project/attrdict/
from pathlib import Path
from datetime import datetime

def search_upwards_for_file(filename):
    """Search in the current directory and all directories above it 
    for a file of a particular name.
    Arguments:
    ---------
    filename :: string, the filename to look for.
    Returns
    -------
    pathlib.Path, the location of the first file found or
    None, if none was found
    """
    d = Path.cwd()
    root = Path(d.root)
    while d != root:
        attempt = d / filename
        if attempt.exists():
            return attempt
        d = d.parent
    return None

def create_partition_key(execution_date:datetime) -> str:
    """
        ARGUMENT
            partition_date: date of airflow task start date. eg 2020-01-01 12:00
    """
    day = '{:02d}'.format(execution_date.day)
    month = '{:02d}'.format(execution_date.month)
    year = execution_date.year
    return f"year={year}/month={month}/day={day}/"

class Config():

    CONFIG_NAME = 'config.yml'
    CONFIG_PATH_AWS = f'/app/{CONFIG_NAME}'
    REQUIRED_GENERAL_KEYS = ['pipeline_name', 'pipeline_bucket', 'pipeline_type', 'prefix_env', 'prefix_logs', 'prefix_raw', 'prefix_processed', 'tasks']

    def __init__(self, **kwargs):
        self.env = kwargs.get('env')
        self.pipeline_type = kwargs.get('type')
        self.airflow_execution_date = kwargs.get('airflow_execution_date')
        self.task = kwargs.get('task')
        self.event_bucket = kwargs.get('event_bucket')
        self.event_prefix = kwargs.get('event_prefix')
        self.file = AttrDict(self._read_config_yaml())
        self.general = self._parse_general()
        self.source = self._parse_source()
        self.transform = self._parse_transform()
        self.target = self._parse_target()

    def _read_config_yaml(self) -> dict:
        file_path = search_upwards_for_file(Config.CONFIG_NAME)
        if not file_path:
            try: 
               file_path = Config.CONFIG_PATH_AWS
            except Exception as e:
                logging.error(f'Config file {Config.CONFIG_NAME} not found') 
                sys.exit(1)
        try:
            with open(file_path) as file:
                config = yaml.safe_load(file)
        except EnvironmentError:
            logging.error('Issue loading config file')
            
        return config

    def _parse_general(self) -> dict:
        """
            General section is valid for all environments
        """
        if self.file.general:
            general = self.file.general
            general['prefix_env'] = getattr(general.prefix_env, self.env) 
            
            # check of all required keys are set up in config, are <> null/empty and have correct type
            for i in Config.REQUIRED_GENERAL_KEYS:
                try:
                    general[i]
                except KeyError as e:
                    logging.error(f'Key \'{i}\' not found in config.yml')
                assert general[i], f"key general.{i} is empty or NULL"
                if i in  ['tasks', 'pipeline_type']:
                    assert isinstance(general[i], list), f"key general.{i} is not a list"
                else: 
                    assert isinstance(general[i], str), f"key general.{i} is not a str" 
        
            # add additional 'calculated' fields to config 
            common_prefix = f'{general.prefix_env}/{general.pipeline_name}/{self.pipeline_type}'
            general['partition_raw'] = f'{common_prefix}/{general.prefix_raw}/{self.task}/{create_partition_key(self.airflow_execution_date)}'
            general['partition_processed'] = f'{common_prefix}/{general.prefix_processed}/{self.task}/{create_partition_key(self.airflow_execution_date)}'
            general['logs'] = f'{general.prefix_logs}/{common_prefix}/{self.task}/{create_partition_key(self.airflow_execution_date)}{datetime.now()}.log'
            return general
        logging.error('Mandatory general section not found')
        sys.exit(1)

    def _parse_source(self) -> dict:
        if self.file.source:
            src = self.file.source

            #### source LYNX
            if src.type == 'db' and src.name =='lynx':

                # required fields
                src['lynx']['schema'] = getattr(src.lynx.schema, self.env)
                src['lynx']['table'] = getattr(src.lynx.table, self.task)
                src['lynx']['extract_size'] = getattr(src.lynx.extract_size, self.task)

            #### source API
            if src.type == 'api':

                # required fields
                src['api']['endpoint']['name'] = getattr(src.api.endpoint.name, self.task)
                src['api']['endpoint']['type'] = getattr(src.api.endpoint.type, self.task)

                if src.api.input.type == 'db' and src.api.input.name =='redshift':

                    # required fields
                    src['api']['input']['redshift']['iam_role'] = getattr(src.api.input.redshift.iam_role, self.env)
                    src['api']['input']['redshift']['query'] = getattr(src.api.input.redshift.query, self.task)

            return src
        logging.error('Mandatory source section not found')
        sys.exit(1)

    def _parse_transform(self) -> dict:
        if self.file.transform:
            transform = self.file.transform

            ## to refactor - we did quick fix here 
            try: 
                if transform.convert_float_to_int:
                    val = getattr(transform.convert_float_to_int, self.task, None)
                    transform['convert_float_to_int'] = list(val) if val else val   
            except Exception:
                ...
            try:  
                if transform.convert_old_timestamps:
                    val = getattr(transform.convert_old_timestamps, self.task, None)
                    transform['convert_old_timestamps'] = list(val) if val else val
            except Exception:
                ...
            return transform
        logging.error('Mandatory transform section not found')
        sys.exit(1)

    def _parse_target(self) -> dict:  # sourcery skip: extract-method
        if self.file.target:
            target = self.file.target
            
            #### Target Redshift
            if target.type == 'db' and target.name == 'redshift':
                
                # required fields
                target['redshift']['iam_role'] = getattr(target.redshift.iam_role, self.env)
                target['redshift']['schema'] = getattr(target.redshift.schema, self.env)
                target['redshift']['table']= getattr(target.redshift.table, self.task)
                target['redshift']['insert_type']= getattr(target.redshift.insert_type, self.task)[self.pipeline_type]

                if target.redshift.insert_type == 'upsert':   # in case of upsert task, upsert keys need to be specified
                    target['redshift']['upsert_keys'] = list(getattr(target.redshift.upsert_keys, self.task))
                else:
                    target['redshift']['upsert_keys'] = None

            return target
        logging.error('Mandatory target section not found')
        sys.exit(1)








import psycopg2
import os, sys
import logging
import pandas as pd
import json
from io import BytesIO
import boto3
from time import time
from typing import List
import uuid

class rsClient():

    def __init__(self, env, rs_iam_role):
        """
            When deploying, environment variables are injected depending on ENV via entrypoint.sh.
            For local development, we fetch environemnt variables from local enviroment where we have variables for local, dev and prd, hence the except statement.

            ! since we don't have a local redshift cluster we load into redshift dev 
        """
        try:
            dbname=os.environ['REDSHIFT_DB']
            user = os.environ['REDSHIFT_UNAME']
            password = os.environ['REDSHIFT_PASSWORD']
            host = os.environ['REDSHIFT_HOST']
            port = os.environ['REDSHIFT_PORT']
        except Exception: 
            dbname=os.environ[f'REDSHIFT_DB_{env.upper()}']
            user = os.environ[f'REDSHIFT_UNAME_{env.upper()}']
            password = os.environ[f'REDSHIFT_PASSWORD_{env.upper()}']
            host = os.environ[f'REDSHIFT_HOST_{env.upper()}']
            port = os.environ[f'REDSHIFT_PORT_{env.upper()}']
        try:
            logging.info('Establishing connection to redshift via psycopg2')
            self.conn = psycopg2.connect(dbname=dbname,user = user,password = password,host = host, port = port)
            logging.info('Redshift Connection succesfully established')
        except psycopg2.DatabaseError as error:
            logging.error('Issue connecting to redshift. Exiting with status 1')
            logging.error(error)
            sys.exit(1)
        self.rs_iam_role = rs_iam_role

    def _enable_autocommit(self):
        self.conn.set_session(autocommit=True)
        return

    def _disable_autocommit(self):
        self.conn.set_session(autocommit=False)
        return

    def _query(self, sql):
        with self.conn.cursor() as cur:
            cur.execute(sql)
            self.conn.commit()
        return 

    def _create_stage_like_target(self,rs_target: str, rs_stage:str, drop_sort_key:bool = False) -> None:
        """
            Checks if staging table exists and drops it. Creates new staging table like target table.
            Drops meta_loading_date_utc column since it is not possible to insert with this column which is populated by database system date when copying stage into target
            In case meta_loading_date_utc has a sort key on target, this sort key needs to be removed first before being able to drop the column. Use drop_sort_key=True in this case
        """
        self._query(f"""DROP TABLE IF EXISTS {rs_stage}""")
        self._query(f"""CREATE TABLE {rs_stage} (LIKE {rs_target})""")
        self._enable_autocommit() 
        if drop_sort_key:
            self._query(f"""ALTER TABLE {rs_stage} ALTER SORTKEY none """)
        self._query(f"""ALTER TABLE {rs_stage} DROP COLUMN meta_loading_date_utc""")
        self._disable_autocommit()
        return

    def _copy_parquet_into_stage(self, files:List[str], rs_stage:str) -> None:
        for file in files:
            try:
                dml = f"""
                    COPY {rs_stage}
                    FROM '{file}'
                    IAM_ROLE '{self.rs_iam_role}'
                    FORMAT AS PARQUET
                """
                self._query(dml)
            except Exception as e:
                logging.error(f'Issue copying {file} into staging table. Exiting')
                logging.error(f'message: {e}', exc_info=True)
                sys.exit(1)
        logging.info('Succesfully loaded all files in temp staging table')
        return
   
    #### write to redshift #########################################################################################################################
    ##############################################################################################################################################
    
    def upsert(self, files:List[str], rs_target_schema:str, rs_target_table:str, upsert_keys:List[str], drop_sort_key=False, **kwargs) -> None:
        """
            ARGUMENTS
                files: list of S3 URI'S
                rs_target_schema: redshift target schema eg. 'ingest'
                rs_target_table: redshift target table eg. 'cogenius_xxx'
                upsert_keys: list of fields to match records between the source (ie. stage) and target, in order to identify which records already exist
                drop_sort_key: optional, needs to be set to True in case target table has a sort key
                            
            RETURNS
                None

            ADDITIONAL INFO
                List of upsert keys can be considered a composite key. If the composite key already exists within the target table, the associated record
                will be deleted and replaced by the new record with the same composite key. 
        """
        stage_uuid = str(uuid.uuid4()).replace('-','')
        rs_target = f'{rs_target_schema}.{rs_target_table}'
        rs_stage = f'{rs_target}_TEMP_{stage_uuid}' 
        self._create_stage_like_target(rs_target, rs_stage, drop_sort_key=drop_sort_key)
        self._copy_parquet_into_stage(files, rs_stage)
        self._upsert(rs_target, rs_stage, upsert_keys)
        return

    def full_refresh(self,files, rs_target_schema, rs_target_table, drop_sort_key=False, **kwargs):
        """
            ARGUMENTS
                files: list of parquet files (eg. ['s3://bucket/folder/sub/file.parquet', ...])
                rs_target_schema: redshift target schema (eg. 'ingest)
                rs_target_table: redshift target table (eg. 'cogenius_xxx')
                drop_sort_key: optional, needs to be set to True in case target table has a sort key
                            
            RETURNS
                None

            ADDITIONAL INFO
                The target table will be truncated (ie we're using delete since truncate commits automatically)
                and stage will be copied into target
        """
        rs_target = f'{rs_target_schema}.{rs_target_table}'
        rs_stage = f'{rs_target}_TEMP_STAGE'
        self._create_stage_like_target(rs_target, rs_stage, drop_sort_key=drop_sort_key)
        self._copy_parquet_into_stage(files, rs_stage)
        self._full_refresh(rs_target, rs_stage)
        return

    def append(self, files: list, rs_target_schema:str, rs_target_table:str, drop_sort_key:bool = False, **kwargs) -> None:
        """
            ARGUMENTS
                files: list of parquet files (eg. ['s3://bucket/folder/sub/file.parquet', ...])
                rs_target_schema: redshift target schema (eg. 'ingest)
                rs_target_table: redshift target table (eg. 'dnb_continual')
                drop_sort_key: optional, needs to be set to True in case target table has a sort key
                            
            RETURNS
                None

            ADDITIONAL INFO
                Records will be appended via insert into
        """
        rs_target = f'{rs_target_schema}.{rs_target_table}'
        rs_stage = f'{rs_target}_TEMP_STAGE'
        self._create_stage_like_target(rs_target, rs_stage, drop_sort_key=drop_sort_key)
        self._copy_parquet_into_stage(files, rs_stage)
        self._append(rs_target, rs_stage)
        return

    def _upsert(self, rs_target:str, rs_stage:str, upsert_keys:List[str]) -> None:
        """
            Upserts from rs_stage into rs_target table. 
            
            1. deletes records in target if upsert key exists in stage
            2. inserts stage into target
            3. drops stage
        """
        where_condition_target = ''
        counter = 0
        for col in upsert_keys:
            if counter == 0:
                line_target = f'{rs_target}.{col} = {rs_stage}.{col}'
            else:
                line_target = f'AND {rs_target}.{col} = {rs_stage}.{col}'
            where_condition_target += str(line_target)
            counter =+ 1
        try:
            self._query(f"""
                    begin transaction;

                    DELETE FROM {rs_target} 
                    USING {rs_stage} 
                    WHERE {where_condition_target} 
                    ;
                    
                    INSERT INTO {rs_target}
                    SELECT *
                    FROM {rs_stage};

                    DROP TABLE {rs_stage};
                    end transaction;
                """)
            logging.info(f'UPSERT succesfull for {rs_stage}')
        except Exception as e:
                logging.error('Issue UPSERTING stage into target. Exiting...')
                logging.error(f'message: {e}', exc_info=True)
                sys.exit(1)

    def _full_refresh(self, rs_target, rs_stage):
        """
            Deletes all data in target table followed by inserting all data from stage into target. 
        """
        # delete is slow but it is not possible to do a truncate within an atomic transaction
        try:
            self._query(f"""
                    begin transaction;

                    DELETE FROM {rs_target} 
                    ;
                    
                    INSERT INTO {rs_target}
                    SELECT *
                    FROM {rs_stage};

                    DROP TABLE {rs_stage};
                    end transaction;
                """)
            logging.info(f'FULL REFRESH succesfull for {rs_target}')
        except Exception as e:
                logging.error('Issue REFRESHING target. Exiting...')
                logging.error(f'message: {e}', exc_info=True)
                sys.exit(1)
    
    def _append(self, rs_target, rs_stage):
        """
            Appends all data from stage into table target. 
        """
        try:
            self._query(f"""
                    begin transaction;
                    
                    INSERT INTO {rs_target}
                    SELECT *
                    FROM {rs_stage};

                    DROP TABLE {rs_stage};
                    end transaction;
                """)
            logging.info(f'Append succesfull for {rs_target}')
        except Exception as e:
                logging.error('Issue appending stage into target. Exiting...')
                logging.error(f'message: {e}', exc_info=True)
                sys.exit(1)

    def insert(self, insert_type: str, files: List[str], target_schema: str, target_table: str, upsert_keys: List[str] = None, drop_sort_key=False ) -> None:
        """ 
            Insert a list of parquet files into redshift target_schema.target_table. 

            Insert options are append, overwrite and upsert.
            In case a sort key exists on "meta_loading_date_utc" set drop_sort_key = True. 
            In case of upsert, specify a list of upsert columns.

            Files are copied file by file into a staging table which is a copy of the target table suffixed with _TEMP_{UUID}. 
           
            UPSERT
            
                1. deletes records in target if upsert key exists in stage
                2. inserts stage into target
                3. drops stage

            APPEND

                1. inserts all data from stage into target. 

            OVERWRITE

                1. deletes all data in target table 
                2. inserts all data from stage into target. 

            TO DO
            
            1. copy a prefix to stage instead of file by file in order to enable parallelization
            2. refactor append, full_refresh, .. methods. Rename them 
            3. drop sort key kwarg should be removed. We should automatically handle the scenario where there is a sort key on meta_loading_date


            GOOD TO KNOW
    
            This is a 'new' method encapsulating the insert methods. For backward compatibility we retain the old function names and access then via an insert_map. 
            Ideally these names should be changed and we should refactor so we can simply do:

            insert = f"insert_{insert_type}"
            if hasattr(self, insert) and callable(func := getattr(self, insert)):
                return func(*args, **kwargs)

        """
        insert_map = {
            'append': self.append,
            'overwrite': self.full_refresh,
            'upsert': self.upsert
        }
        insert_func = insert_map[insert_type]
        insert_func(files, target_schema, target_table, upsert_keys=upsert_keys, drop_sort_key=drop_sort_key)
        return

    def load_copy_csv(self, s3_bucket, s3_prefix, rs_target, iam_role, delimiter, columns=None, timeformat='YYYY-MM-DDTHH:MI:SS'):

        # UNTESTEd

        # sourcery skip: none-compare
        """
            s3_bucket: bucket
            s3_prefix: path/path/path/
            rs_target: target_schema.target_table
            env: dev or prd
            columns: ['col1', 'col2', ...]

            ColumnA in your data file will map to col1 in rs, columnB in your file will map to col2 in rs.
            Columns >= len(columns in file). You can use this to create a RS table with X cols, a file with X-1 cols and have 1 col in redshift 
            defaulting to a value, for example a loading date (ie. loading date TIMESTAMP default sysdate)
        """
        path = 's3://'+s3_bucket+'/'+s3_prefix
        if columns == None:
            dml = f"""
                COPY {rs_target}
                FROM '{path}'
                CVS
                IAM_ROLE '{iam_role}'
                delimiter '{delimiter}'
                timeformat '{timeformat}'
                EMPTYASNULL
            """
        else: 
            dml = f"""
                COPY {rs_target} ('{",".join(columns)}')
                FROM '{path}'
                CSV
                IAM_ROLE '{iam_role}'
                delimiter '{delimiter}'
                timeformat '{timeformat}'
                EMPTYASNULL
            """
        self._query(dml)
        return

    #### read redshift  ##########################################################################################################################
    ##############################################################################################################################################

    def rs_to_df(self, dml):
        return pd.read_sql_query(dml, self.conn)

    def rs_fetch_single_val(self, query):
        result = None
        with self.conn.cursor() as cur:
            cur.execute(query)
            result = cur.fetchone()
            self.conn.commit()
        return result[0]

    def rs_column_to_api_query_param_list(self, rs_source_schema:str, rs_source_table:str, rs_source_column:str, api_query_params:str):
        """
            ARGUMENTS:
                rs_source_schema: eg. ingest
                rs_source_table: eg. table
                rs_source_column: eg column1
                api_query_params: eg objectId in case API endpoint is url/objectId?value1 where value1 is extracted from redshift ingest.table.column1
            RETURNS
                list of stringified json [{'objectId': column1[value1]}, {'objectId': column1[value2]}]

            ADDITIONAL INFO:
                For several APIs we need a redshift column as input. For each cell value we'll fetch a data object from an API via query parameters.
                Hence, we create a list of stringified dictionaries which can be readily passed into a fetch function containing the relevant cell value and
                query parameter.
        """
        dml = f"""
            SELECT "{rs_source_column}" from {rs_source_schema}.{rs_source_table} where "{rs_source_column}" is not null
        """
        df = self.rs_to_df(dml)
        # sometimes we have floats in redshift that should be integers, hence we need to cast to integers.
        float_col = df.select_dtypes(include=['float64']) 
        for col in float_col.columns.values:
            df[col] = df[col].astype('int64')
        input_list= df[rs_source_column].tolist()
        api_queries = [json.dumps({api_query_params: x}) for x in input_list]
        return api_queries

    def rs_column_to_list(self, rs_source_schema:str, rs_source_table:str, rs_source_column:str):
        dml = f"""
            SELECT "{rs_source_column}" from {rs_source_schema}.{rs_source_table} where "{rs_source_column}" is not null
        """
        df = self.rs_to_df(dml)
        # sometimes we have floats in redshift that should be integers, hence we need to cast to integers.
        float_col = df.select_dtypes(include=['float64']) 
        for col in float_col.columns.values:
            df[col] = df[col].astype('int64')
        path_params= df[rs_source_column].tolist()
        return path_params

    def rs_query_to_list(self, query: str, rs_source_column: str):
        df = self.rs_to_df(query)
          # sometimes we have floats in redshift that should be integers, hence we need to cast to integers.
        float_col = df.select_dtypes(include=['float64']) 
        for col in float_col.columns.values:
            df[col] = df[col].astype('int64')
        path_params= df[rs_source_column].tolist()
        return path_params

    def stream_to_s3_parquet(self, query:str, batch_size:int, s3_bucket:str, s3_prefix:str, s3_filename:str) -> None:
        # sourcery skip: identity-comprehension
        """
            Streams the results of a sql query to parquet files on s3 with 'batch_size' nbr of rows per file. 
            Output files have the following key:
                {s3_bucket}{s3_prefix}{timestamp}_{filename}.parquet        
        """
        for rows, cols in self._stream_results(query, batch_size):
            get_data = [x for x in rows]
            col_names = [y[0] for y in cols]
            df = pd.DataFrame(get_data)
            df.columns = col_names
            self._df_to_parquet_s3(df, s3_bucket, s3_prefix, s3_filename)
        return
      
    def _stream_results(self, query:str, batch_size: int):
        cursor = self.conn.cursor()
        cursor.itersize = batch_size
        cursor.execute(query)
        while True:
            rows = cursor.fetchmany(batch_size)
            cols = cursor.description
            if not rows:
                break
            yield rows, cols
            
    def _df_to_parquet_s3(self, df:pd.DataFrame, s3_bucket: str, s3_prefix: str, file_name:str):
        parquet_buffer = BytesIO()
        df.to_parquet(parquet_buffer, index=False, allow_truncated_timestamps=True)
        s3 = boto3.resource('s3')
        timestamp = round(time(), 4)
        s3_key = f"{s3_prefix}{timestamp}_{file_name}.parquet"
        s3.Object(s3_bucket, s3_key).put(Body=parquet_buffer.getvalue())
        logging.info(f'Stored {s3_key} on s3 {s3_bucket}')
        del df
        return



import requests
import json
import random
from time import sleep
import logging
import ast
import sys, os
from typing import Dict, List
import pandas as pd

class basicAPI():

    def __init__(self, url, auth_type = 'api_key', **kwargs ):
        if url.endswith("/"):
            None
        else:
            url = f'{url}/'
        self.API_URL= url
        self.auth_type = auth_type
        self.api_key = kwargs.get('api_key')
        if self.auth_type == 'api_key':
            self.headers = {
                'X-API-KEY' : self.api_key,
                'Accept': 'application/json'
            } 
        self.session = requests.Session()
        self.session.headers.update(self.headers)
        return 

    def fetch_query_params(self, query_params, endpoint, retries=3, backoff_in_seconds=1):
        """
            ARGUMENTS
                query_params: list of stringified dictionary, eg. '{'ean': 12345}'
                endpoint: endpoint part within API endpoint eg. self.url/V1/endpoint?ean=12345

            RETURNS
                response_dictionary or a runtimeError
            
        """
        x = 0
        while True:
            try:
                return self._fetch_query_params(endpoint, query_params)
            except RuntimeError:
                if x == int(retries)-1:
                    raise RuntimeError(f'No 200 response after {retries} tries for query: {json.loads(query_params)}')
                else:
                    # logging.warning(f'Trying again ... query: {json.loads(query_params)}')
                    sleep_duration = (int(backoff_in_seconds) * 2 ** x + random.uniform(0, 1))
                    sleep(sleep_duration)
                    x += 1

    def fetch_path_params(self, path_params, endpoint, retries=3, backoff_in_seconds=1):
        x = 0
        while True:
            try:
                return self._fetch_path_params(endpoint, path_params)
            except RuntimeError:
                if x == int(retries)-1:
                    raise RuntimeError(f'No 200 response after {retries} tries for {endpoint}/{path_params}')
                else:
                    # logging.warning(f'Trying again ... {endpoint}/{path_params}')
                    sleep_duration = (int(backoff_in_seconds) * 2 ** x + random.uniform(0, 1))
                    sleep(sleep_duration)
                    x += 1

    def _fetch_query_params(self, endpoint, query_params):
        url_endpoint = f'{self.API_URL}{endpoint}'
        response = self.session.get(url_endpoint, params = json.loads(query_params))
        if response.status_code != 200:
            raise RuntimeError(f' {response.status_code} response for {json.loads(query_params)}')
        response_dict=json.loads(response.text)
        # we need to add the query parameter to the response in order to link tables. We add query parameter as api_id
        # for adresmatch, we add straatnaam, huisnr, gemeentenaam en postcode
        d = ast.literal_eval(query_params)
        if len(d)==1:
            (k, v), = d.items()
            response_dict['api_id'] = v
        else:
            for k,v in d.items():
                response_dict[k] = v
        return response_dict

    def _fetch_path_params(self, endpoint, path_params):
        url_endpoint = f'{self.API_URL}{endpoint}/{path_params}'
        response = self.session.get(url_endpoint)
        if response.status_code != 200:
            raise RuntimeError(f' {response.status_code} response for {endpoint}/{path_params}')
        response_dict=json.loads(response.text)
        # we need to add the path parameter to the response
        response_dict['api_id']=path_params
        return response_dict

    def close_session(self):
        try:
            self.session.close()
            logging.info("Closed API session")
        except Exception as e:
            logging.error("Error closing API session")


class CogeniusAPI():
    def __init__(self, endpoint):
        try:
            COGENIUS_API_KEY=os.environ[f'COGENIUS_API_KEY']
            COGENIUS_API_ENDPOINT = os.environ[f'COGENIUS_API_ENDPOINT']
            logging.info("Found cogenius credentials in os.environ")
        except Exception as e:
            logging.error("No cogenius credentials found in os.environ. Exiting...")
            logging.error(e)
            sys.exit(1)
        self.endpoint_url = f'{COGENIUS_API_ENDPOINT}{endpoint}'
        logging.info(self.endpoint_url)
        self.headers = {
            'X-API-KEY' : COGENIUS_API_KEY,
            'Accept': 'application/json'
        } 
        self.session = requests.Session()
        self.session.headers.update(self.headers)
        return 

    def fetch(self, query, retries=3, backoff_in_seconds=1):
        """
            returns response_dictionary or a runtimeError
            query: stringified dictionary, eg. '{'ean': 12345}'
        """
        x = 0
        while True:
            try:
                return self._fetch(query)
            except RuntimeError:
                if x == int(retries)-1:
                    raise RuntimeError(f'No 200 response after {retries} tries for query: {json.loads(query)}')
                else:
                    logging.warning(f'Trying again ... query: {json.loads(query)}')
                    sleep_duration = (int(backoff_in_seconds) * 2 ** x + random.uniform(0, 1))
                    sleep(sleep_duration)
                    x += 1

    def _fetch(self, query):
        response = self.session.get(self.endpoint_url, params = json.loads(query))
        if response.status_code != 200:
            raise RuntimeError(f' {response.status_code} response for {json.loads(query)}')
        response_dict=json.loads(response.text)
        return response_dict

    def close_session(self):
        try:
            self.session.close()
            logging.info("Closed cogenius API session")
        except Exception as e:
            logging.error("Error closing cogenius session")
        

class basisregisterAPI():

    def __init__(self, url, auth_type = 'api_key', **kwargs ):
        if url.endswith("/"):
            None
        else:
            url = f'{url}/'
        self.API_URL= url
        self.auth_type = auth_type
        self.api_key = kwargs.get('api_key')
        if self.auth_type == 'api_key':
            self.headers = {
                'X-API-KEY' : self.api_key,
                'Accept': 'application/json'
            } 
        self.session = requests.Session()
        self.session.headers.update(self.headers)
        return 

    def create_api_params_from_df(self,  df: pd.DataFrame) -> List[str]:
        """
            Method to create a list of json strings containing api params from a df of param columns.

            Given df: 
            +-----+-----+
            | col1| col2|
            +-----+-----+
            |   z1|   a1|
            |   z1|   b2|
            +-----+-----+

            Returns:
            ['{'col1': 'z1', 'col2': 'a1'}', '{...}']
                 

        """
        # we transform df to list of json dictionaries and remove keys if value is None
        api_params = [json.dumps({k: v for k, v in rec.items() if v is not None}) for rec in df.to_dict('records')]   
        return api_params

    def fetch(self, params: str, endpoint:str , fetch_type: str, retries=3, backoff_in_seconds=1, **kwargs) -> Dict:
        """
            ARGUMENTS
                params: json string of params, eg. '{'param1': 12345, 'param2': 'abc'}' 
                endpoint: endpoint suffix eg. url/{endpoint}
                fetch_type: path or query. In case of query, params are passed as params. In case of path, the suffix /param1_value/param2_value/... is added to the endpoint

            RETURNS
                api response object enriched with the input params prefixed with api_param_: eg api_param_param1 : 12345
        """
        assert fetch_type in {'query', 'path'}, f"Invalid fetch type {fetch_type}"
        x = 0
        while True:
            try:
                return self._fetch(params, endpoint, fetch_type, **kwargs)
            except RuntimeError as e:
                if x == int(retries)-1:
                    raise RuntimeError(
                        f'No 200 response after {retries} tries for query: {json.loads(params)}'
                    ) from e
                sleep_duration = (int(backoff_in_seconds) * 2 ** x + random.uniform(0, 1))
                sleep(sleep_duration)
                x += 1

    def _fetch(self, params: str, endpoint:str, fetch_type:str, **kwargs) -> Dict:
        if fetch_type == 'query':
            url_endpoint = f'{self.API_URL}{endpoint}'
            res = self.session.get(url_endpoint, params = json.loads(params))
        if fetch_type == 'path':
            params_suffix = '/'.join(list(ast.literal_eval(params).values()))
            url_endpoint = f'{self.API_URL}{endpoint}/{params_suffix}'
            res = self.session.get(url_endpoint)
        res_dic=json.loads(res.text)
        d = ast.literal_eval(params)
        for k,v in d.items():
            res_dic[f'api_param_{k}'] = v
        if res.status_code != 200:
            raise RuntimeError(f' {res.status_code} response for API call with {fetch_type} parameters: {json.loads(params)}')
        return res_dic

    def close_session(self):
        try:
            self.session.close()
            logging.info("Closed API session")
        except Exception as e:
            logging.error("Error closing API session")
      



import contextlib
import pandas as pd
import shutil, os
import logging
from datetime import datetime, timezone
from time import time
from functools import reduce
import psutil
import sys
import json
from typing import List, Dict
from operator import itemgetter
import uuid

from pyspark.sql import SparkSession
from pyspark.sql import DataFrame as SparkDataFrame
import pyspark.sql.functions as F
from pyspark.sql.functions import udf
from pyspark.sql.types import StructType, StructField, StringType, IntegerType, BooleanType, TimestampType, DoubleType, DecimalType, ArrayType, BinaryType, LongType
import boto3

from . import s3_client

class sparkClient():
    """
        env: local, dev or prd
    """

    ### create session ###########################################################################################################################
    ##############################################################################################################################################
    def __init__(self, env:str, **kwargs):
        self.env = env
        self.spark = self._create_spark_session()
        return None

    def _get_temp_batch_credentials(self ):
        # we cannot fetch sts temp credentials via an assumed role, hence, we get the current credentials of the assumed role
        # https://stackoverflow.com/questions/36287720/boto3-get-credentials-dynamically
        session = boto3.Session()
        credentials = session.get_credentials()
        credentials = credentials.get_frozen_credentials()
        access_key = credentials[0]
        secret_key = credentials[1]
        session_token = credentials[2]
        return access_key, secret_key, session_token

    def _create_spark_session(self):
        """
           Create a spark session on a local cluster (ie. single node) using nbr_of_cores cores.

           In case spark runs on AWS batch, we need to use temporary AWS credentials in order to read from 
           S3. In case of local dev we can use default AWS profile. 
        """
         # https://stackoverflow.com/questions/50891509/apache-spark-codegen-stage-grows-beyond-64-kb 
        nbr_cores = self._get_nbr_cores()

        try:
            if self.env == 'local':
                logging.info('Building spark session w ProfileCredentialsProvider for S3 access')
                spark = SparkSession.builder.master(f"local[{nbr_cores}]") \
                            .appName(f'Spark_{str(uuid.uuid4())}') \
                            .config('spark.sql.codegen.wholeStage', 'false') \
                            .config("spark.sql.session.timeZone", "UTC") \
                            .config('spark.jars.packages', 'org.apache.hadoop:hadoop-aws:3.2.0')\
                            .config("spark.sql.legacy.parquet.datetimeRebaseModeInRead", "LEGACY") \
                            .config("fs.s3a.aws.credentials.provider","com.amazonaws.auth.profile.ProfileCredentialsProvider")\
                            .getOrCreate()

            else:
                access_key, secret_key, session_token = self._get_temp_batch_credentials()
                # https://stackoverflow.com/questions/54223242/aws-access-s3-from-spark-using-iam-role?noredirect=1&lq=1
                logging.info('Building spark session w TemporaryAWSCredentialsProvider for S3 access')
                spark = SparkSession.builder.master(f"local[{nbr_cores}]") \
                            .appName(f'Spark_{str(uuid.uuid4())}') \
                            .config('spark.sql.codegen.wholeStage', 'false') \
                            .config("spark.sql.session.timeZone", "UTC") \
                            .config('spark.jars.packages', 'org.apache.hadoop:hadoop-aws:3.2.0')\
                            .config("spark.sql.legacy.parquet.datetimeRebaseModeInRead", "LEGACY") \
                            .config('fs.s3a.aws.credentials.provider', 'org.apache.hadoop.fs.s3a.TemporaryAWSCredentialsProvider')\
                            .config('fs.s3a.access.key', access_key)\
                            .config('fs.s3a.secret.key', secret_key)\
                            .config('fs.s3a.session.token', session_token)\
                            .getOrCreate()  

            spark.sparkContext.setLogLevel("ERROR")
            logging.info("Spark Session created")
        except Exception as e:
            logging.error("Issue creating Spark Session. Exiting...")
            logging.error(e)
            sys.exit(1)
        return spark

    @staticmethod
    def _get_nbr_cores() -> int:
       return psutil.cpu_count(logical = False)

    ### read into spark df  ######################################################################################################################
    ##############################################################################################################################################
    def read_parquet(self, bucket: str, prefix:str) -> SparkDataFrame:
        """
            Read a prefix containing parquet files into a df. 

            In case of reading raw parquet files from lynx queries, apply enforce_schema since we stream lynx in parts. As a result, parquet files 
            might have slightly different schemas. For example, columns which contain only nulls in one file but values in another
        """
        key = f's3a://{bucket}/{prefix}'
        return self.spark.read.parquet(key)

    def read_single_csv_file(self, file:str, schema: Dict[str, StructField] = None, header: str = "true", delimiter: str = ";") -> SparkDataFrame:
        """
            file: complete s3 URI including s3://
        """
        file = file.replace('s3://', 's3a://')
        try: 
            df = self.spark.read.format("csv").option("header",header).option("delimiter", delimiter).load(file)
            logging.info(f"Read {file} in spark DF")
            if schema:
                new_df = self.enforce_schema(df, schema)
                logging.info("Enforced schema")
            else:
                new_df = df
        except Exception as e:
            logging.error("Issue reading parquet Exiting...")
            logging.error(e)
            sys.exit(1)
        return new_df

    def read_single_parquet_file(self, file: str, schema: Dict[str, StructField] = None) -> SparkDataFrame:
        """ 
            file: complete s3 URI including s3://
        """
        file = file.replace('s3://', 's3a://')
        try: 
            df = self.spark.read.format("parquet").load(file)
            logging.info(f"Read {file} in spark DF")
            if schema:
                new_df = self.enforce_schema(df, schema)
                logging.info("Enforced schema")
            else:
                new_df = df
        except Exception as e:
            logging.error("Issue reading parquet Exiting...")
            logging.error(e)
            sys.exit(1)
        return new_df

    def read_multiple_csv_files(self, files: List[str], schema: Dict[str, StructField] = None, header: str = "true", delimiter: str = ";") -> SparkDataFrame:
        """
            files: list of complete s3 URI including s3://
        """   
        try:
            list_of_dfs = []
            for file in files:
                file = file.replace('s3://', 's3a://')
                df = self.spark.read.format("csv").option("header",header).option("delimiter", delimiter).load(file)
                if schema:
                    new_df = self.enforce_schema(df, schema)
                    list_of_dfs.append(new_df)
                else:
                    list_of_dfs.append(df)
            united_df = reduce(self._unite_dfs, list_of_dfs)
            logging.info("Succesfully read all files in a spark DF")
        except Exception as e:
            logging.error("Issue reading parquet files and unioning them in spark DF Exiting...")
            logging.error(e)
            sys.exit(1)
        return united_df
    
    def read_multiple_parquet_files(self, files: List[str], schema: Dict[str, StructField] = None) -> SparkDataFrame:
        """
            files: list of complete s3 URI including s3://
        """   
        try:
            list_of_dfs = []
            for file in files:
                file = file.replace('s3://', 's3a://')
                df = self.spark.read.format("parquet").load(file)
                if schema:
                    new_df = self.enforce_schema(df, schema)
                    list_of_dfs.append(new_df)
                else:
                    list_of_dfs.append(df)
            united_df = reduce(self._unite_dfs, list_of_dfs)
            logging.info("Succesfully read all files in a spark DF")
        except Exception as e:
            logging.error("Issue reading parquet files and unioning them in spark DF Exiting...")
            logging.error(e)
            sys.exit(1)
        return united_df

    def pandas_df_to_spark_df(self, pandas_df: pd.DataFrame, schema=None) -> SparkDataFrame:
        if schema:
            df_spark = self.spark.createDataFrame(pandas_df, schema=schema)
        else:
            df_spark = self.spark.createDataFrame(pandas_df)
        return df_spark

    #### write spark df  #########################################################################################################################
    ##############################################################################################################################################
    def write_to_parquet(self, spark_df: SparkDataFrame, s3_bucket: str, s3_prefix: str, max_records_per_file = 100000) -> None:
        s3 = s3_client.s3Client(s3_bucket)
        local_path = './data'
        with contextlib.suppress(Exception):
            shutil.rmtree(local_path)
        try:
            spark_df.write.option('maxRecordsPerFile', max_records_per_file).mode('overwrite').parquet(local_path)
            for file in os.listdir(local_path):
                local_file = f'{local_path}/{file}'
                if '.crc' not in file and 'SUCCESS' not in file:
                    s3.upload_local_file(local_file, s3_prefix )
                os.remove(local_file)
            try:
                os.rmdir(local_path)
            except Exception:
                None
            logging.info(f'Succesfully wrote {s3_prefix}')
        except Exception as e:
            logging.error(f'Issue creating parquet file: {s3_prefix}. Exiting...')
            logging.error(e)
            sys.exit(1)
        return 

    @staticmethod
    def spark_df_to_pandas_df(df: SparkDataFrame) -> pd.DataFrame:
        return df.toPandas()
    
    def apply_schema(self, df: SparkDataFrame, schema: StructType) -> SparkDataFrame:
        """
            Method that first reoders the columns to align with the Spark Schema and applies it afterwards
        """
        try:
            df = df.select(StructType.fieldNames(schema))
            df = self.spark.createDataFrame(df.collect(), schema = schema)
            logging.info("Succesfully applied schema")
        except Exception as e:
            logging.error("Issue applying schema. Exiting...")
            logging.error(e)
            sys.exit(1)
        return df


    #### UDF & their methods #####################################################################################################################
    ##############################################################################################################################################
    @staticmethod
    @udf
    def udf_float_to_int(x):
        if x is None:
            res = None
        else:
            try:
                res = int(x)
            except AttributeError:
                res = x
        return res

    @staticmethod
    @udf
    def udf_string_to_int(x):
        if x is None or x =='':
            res = None
        else:
            try:
                # https://stackoverflow.com/questions/1841565/valueerror-invalid-literal-for-int-with-base-10 
                res = int(x)
            except AttributeError:
                res = x
        return res

    @staticmethod
    @udf
    def udf_string_to_timestamp(x):
        if x is None or x =='':
            res = None
        else:
            x = x.split('.')[0]
            date_format = '%Y-%m-%d %H:%M:%S'
            try:
                # https://stackoverflow.com/questions/1841565/valueerror-invalid-literal-for-int-with-base-10 
                res = datetime.strptime(x, date_format)
            except AttributeError:
                res = x
        return str(res)

    @staticmethod
    @udf
    def udf_unicode(x):
        if x is None: 
            res = None
        else:
            try:
                res = x.encode("ascii","ignore")
            except AttributeError:
                res = x
        return res

    @staticmethod
    @udf
    def udf_filter_max_from_array(x:str, sort_field:str) -> any:
        data = json.loads(x)
        if not data:
            return json.dumps(data)
        max_element = max(data, key=itemgetter(sort_field))
        return json.dumps(max_element)

    def convert_float_to_int_int(self, df: SparkDataFrame, cols: list) -> SparkDataFrame:
        """
            When reading from SQL, columns which are INT but contain only NULLS get converted to parquet float columns.
            We create a pyspark schema based on the original SQL schema, hence INT column gets translated into StructField('col', IntegerType(), True).
            In order to apply this pyspark schema, we need to convert the relevant int columns back to int.
        """
        try: 
            new_df = df.select(*[self.udf_float_to_int(column).cast(IntegerType()).alias(column) if column in cols else column for column in df.columns])
            logging.info("Succesfully converted float columns back to int")
        except Exception as e:
            logging.error("Issue converting float columns to int. Exiting...")
            logging.error(e)
            sys.exit(1)
        return new_df

    def convert_string_to_int_int(self, df: SparkDataFrame, cols: list) -> SparkDataFrame:
        """
            Convert list of string cols to IntegerType() by explicit casting
        """
        try: 
            new_df = df.select(*[self.udf_string_to_int(column).cast(IntegerType()).alias(column) if column in cols else column for column in df.columns])
            logging.info("Succesfully converted str columns to int")
        except Exception as e:
            logging.error("Issue converting str columns to int. Exiting...")
            logging.error(e)
            sys.exit(1)
        return new_df

    def convert_string_to_timestamp(self, df: SparkDataFrame, cols: list) -> SparkDataFrame:
        """
            Convert list of string cols to TimestampType() by explicit casting
        """
        try: 
            print('cols')
            print(cols)
            new_df = df.select(*[self.udf_string_to_timestamp(column).cast(TimestampType()).alias(column) if column in cols else column for column in df.columns])
            logging.info("Succesfully converted string columns to timestamp")
        except Exception as e:
            logging.error("Issue converting str columns to timestamp. Exiting...")
            logging.error(e)
            sys.exit(1)
        return new_df

    def convert_version(self, df: SparkDataFrame) -> SparkDataFrame: 
        if 'version' in df.columns:
            try: 
                new_df = df.select(*[self.udf_unicode(column).alias('version') if column == 'version' else column for column in df.columns])
                logging.info("Succesfully converted lynx version column to ascii.")
            except Exception as e:
                logging.error("Issue converting lynx version column.Exiting...")
                logging.error(e)
                sys.exit(1)
        else: new_df = df
        return new_df

    def filter_max_from_array(self, df: SparkDataFrame, spark_array: str, sort_field: str ) -> SparkDataFrame:
        """
            Method to filter an array and only keep the array_element for which max(array_element.sort_field)
        """ 
        temp_col_name = str(uuid.uuid4())
        df = df.withColumn(temp_col_name, self.udf_filter_max_from_array(F.to_json(F.col(spark_array)), F.lit(sort_field)))
        df = df.drop(spark_array)
        df_json_schema = self.spark.read.json(df.rdd.map(lambda row: row[temp_col_name])).schema
        df = df.withColumn(spark_array, F.from_json(F.col(temp_col_name), df_json_schema))
        df = df.drop(temp_col_name)
        return df

    #### static dataframe manipulation methods ###################################################################################################
    ##############################################################################################################################################
    @staticmethod
    def add_metadata(df:SparkDataFrame, file_name = None, partition_date = datetime(2020,1,1)) -> SparkDataFrame:
        """
            partition_date: airflow execution date passed as a datetime.datetime object
        """
        if not file_name: file_name = F.input_file_name()
        # partition_date = str(partition_date.date())
        try: 
            df = df.withColumn('META_file_name', F.lit(file_name)) \
                        .withColumn('META_partition_date', F.lit(partition_date)) \
                        .withColumn('META_processing_date_utc', F.lit(datetime.now(timezone.utc)))
            logging.info("Succesfully added metadata")
        except Exception as e:
            logging.error("Issue adding metadata. Exiting...")
            logging.error(e)
            sys.exit(1)
        return df

    @staticmethod
    def enforce_schema(df: SparkDataFrame, schema: Dict[str, StructField]) -> SparkDataFrame:  
        try:
            new_df = df.select([F.col(c).cast(schema[c].dataType).alias(c) for c in df.columns])
            # new_df = df.select([F.col(c).cast(schema[c]).alias(c) for c in df.columns])
            logging.info("Succesfully applied schema")
        except Exception as e:
            logging.error("Issue enforcing schema. Exiting...")
            logging.error(e)
            sys.exit(1)
        return new_df

    @staticmethod
    def _unite_dfs(df1: SparkDataFrame, df2: SparkDataFrame) -> SparkDataFrame:
        return df1.unionByName(df2)

    @staticmethod
    def reindex_cols(df: SparkDataFrame, columns_order: List[str]) -> SparkDataFrame:
        try:
            new_df = df.select(*columns_order)
            logging.info("Succesfully reindexed columns")
        except Exception as e:
            logging.error("Issue reindexing columns. Exiting...")
            logging.error(e)
            sys.exit(1)
        return new_df

    @staticmethod
    def clean_old_dates(df:SparkDataFrame, cols_to_clean:list) -> SparkDataFrame:
        """
            Spark 3.0 has difficulties working with dates older than 1900-01-01.

            org.apache.spark.SparkUpgradeException: You may get a different result due to the upgrading of Spark 3.0: reading dates before 1582-10-15 or timestamps before 1900-01-01T00:00:00Z from Parquet files can be ambiguous, as the files may be written by Spark 2.x or legacy versions of Hive, which uses a legacy hybrid calendar that is different from Spark 3.0+'s Proleptic Gregorian calendar. See more details in SPARK-31404. You can set spark.sql.legacy.parquet.datetimeRebaseModeInRead to 'LEGACY' to rebase the datetime values w.r.t. the calendar difference during reading. Or set spark.sql.legacy.parquet.datetimeRebaseModeInRead to 'CORRECTED' to read the datetime values as it is.
        
            Adding datetimeRebaseModeInRead still results in 'mktime argument out of range' errors. Hence, we'll manually clean these dates.
        """
        new_df = df.select([
            F.when(F.col(column) < "1900-01-01 00:00:00", F.lit("1900-01-01 00:00:00"))\
            .otherwise(F.col(column))\
            .cast(TimestampType())\
            .alias(column) if column in cols_to_clean else column for column in df.columns
            ]
        )
        return new_df

    @staticmethod
    def nan_to_null(df: SparkDataFrame) -> SparkDataFrame:
        """
            In case you have string columns with NaN strings ( eg after reading pd.read_csv) which you want 
            to load as NULL in redshift, apply this function. 
        """
        for col in df.columns:
            df = df.withColumn(col, F.when(F.col(col) == 'NaN', None).otherwise(F.col(col)))
        return df
    
    @staticmethod
    def flatten_struct(df: SparkDataFrame, flatten_col: StructType) -> SparkDataFrame:
        """     
            Method to flatten a StructType
        """
        cols = df.columns
        cols.remove(flatten_col)
        df = df.select(*cols, f'{flatten_col}.*')
        return df

    @staticmethod
    def filter_empty_from_array(df: SparkDataFrame, spark_array: str, filter_field: str) -> SparkDataFrame:
        """
            Method to filter elements from an array when array_element.filter_field is null
        """
        temp_col_name = str(uuid.uuid4())
        # https://stackoverflow.com/questions/73293720/pyspark-filter-an-array-of-structs-based-on-one-value-in-the-struct
        df = df.withColumn(temp_col_name, F.expr(f"filter({spark_array}, x -> x.{filter_field} is not null)")) 
        df = df.drop(spark_array)
        df = df.withColumnRenamed(temp_col_name, spark_array)
        return df

    @staticmethod
    def explode_column(df: SparkDataFrame, explode_col: str) -> SparkDataFrame:
        """    
            Explodes a column which contain a list of items. The col type can either be a string '['a', 'b']' or an array.

            Given df 
            +------+-------------+------+
            |  col1|         col2|  col3|
            +------+-------------+------+
            |    z1| [a1, b2, c3]|   foo|
            +------+-------------+------+

            explode_columns(df, col2) will result in 

            +-----+-----+-----+
            | col1| col2| col3|
            +-----+-----+-----+
            |   z1|   a1|  foo|
            |   z1|   b2|  foo|
            |   z1|   c3|  foo|
            +-----+-----+-----+
        """
        col_type = dict(df.dtypes)[explode_col]
        temp_col_name = str(uuid.uuid4())
        if col_type == 'string':
            # https://stackoverflow.com/questions/57066797/pyspark-dataframe-split-column-with-multiple-values-into-rows
            df = df.withColumn(temp_col_name, F.explode( F.split( F.regexp_extract( F.regexp_replace(F.col(explode_col), "\s", ""), "^\[(.*)\]$", 1), ",") ) )
            df = df.drop(explode_col)
            df = df.withColumnRenamed(temp_col_name, explode_col)
        if 'array' in col_type:
            df = df.withColumn(temp_col_name, F.explode(explode_col))
            df = df.drop(explode_col)
            df = df.withColumnRenamed(temp_col_name, explode_col)
        return df

    @staticmethod
    def drop_cols(df: SparkDataFrame, list_cols: List[str]) -> SparkDataFrame:
        """
            Method that drops al columns in the list_cols 1 by 1.
        """
        for c in list_cols:
            df = df.drop(c)
        return df


            






import os
import yaml
import inspect
from datetime import date, datetime
import logging
import sys
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
import psutil
import argparse
from typing import Dict

cores = psutil.cpu_count(logical = False)

class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

def get_config():
    """
        RETURN
            config object enabling fetching of config from config.yml via dot notation

        ADDITIONAL INFO
            In docker image we copy config file into current dir, for local development we fetch config from level lower
    """
    caller_file = os.path.dirname(inspect.stack()[1].filename)
    try: 
        path = os.path.join(caller_file, 'config.yml')
        with open(f'{caller_file}/config.yml') as file:
            config = yaml.load(file, Loader=yaml.FullLoader)

    except Exception:
        path = os.path.join(caller_file, '..', 'config.yml')
        with open(path) as file:
            config = yaml.load(file, Loader=yaml.FullLoader)

    config = dotdict(config)
    return config

def create_partition_key(partition_date:str) -> str:
    """
        ARGUMENT
            partition_date: date of airflow task start date. eg 2020-01-01 12:00
        
        RETURN
            year=2020/month=01/day=01/    
    """
    if not isinstance(partition_date, date):
        date_format = "%Y-%m-%d"
        partition_date = datetime.strptime(partition_date, date_format)
    day = '{:02d}'.format(partition_date.day)
    month = '{:02d}'.format(partition_date.month)
    year = partition_date.year
    return f"year={year}/month={month}/day={day}/"

def get_args(args):
    """
        ARGUMENTS
            args: sys.args

        RETURNS
            env, pipeline_type, partition_start, partition_start, partition_key

        ADDITIONAL INFO
            Command line args are passed to python programs via airflow dag - AWS Batchoperator - Overrides - Command  
            These args are then passed to our program via entrypoint.sh via $@

            Env, pipeline type and partition start ( ie. airflow task start date) are mandatory for incremental/full pipelines.
            For incremental pipeline, additionally we require a partition end date. 
            Partition key is also returned which is calculated based on partition_start
    """
    try: 
        print(args)
        env = args[1]
        pipeline_type = args[2]
        assert env in ['local', 'dev', 'prd']
        assert pipeline_type in ['full', 'incremental']
        logging.info(f"Environment: {env}")
        logging.info(f"Pipeline type: {pipeline_type}")
    except AssertionError as e:
        logging.error("Incorrect environment or pipeline type argument passed. Exiting...")
        logging.error(e)
        sys.exit(1)
    except Exception:
        logging.error("No environment argument passed. Exiting...")
        sys.exit(1)

    # fetch airflow partition(s)
    try:
        format = "%Y-%m-%d"
        partition_start = args[3]
        try:
            is_valid_start = bool(datetime.strptime(partition_start, format))
        except:
            is_valid_start = False
        assert is_valid_start
        logging.info(f"Partition start: {partition_start}")
        if pipeline_type == 'incremental':
            partition_end = args[4]
            try:
                is_valid_end = bool(datetime.strptime(partition_end, format))
            except:
                is_valid_end= False
            assert is_valid_end
            logging.info(f"Partition end: {partition_end}")
        else:
            partition_end = None
        partition_key = create_partition_key(partition_start)
    except AssertionError as e:
        logging.error("Incorrect partition date argument passed. Dates must be passed as YYYY-MM-DD Exiting...")
        logging.error(e)
        sys.exit(1)
    except: 
        logging.error("No partition date arguments passed. Exiting")
        logging.info("Pass partition start and partition end date arguments as postional arguments 2 and 3")
        sys.exit(1)
    return env, pipeline_type, partition_start, partition_end, partition_key

def chunker(seq, size):
    """
        ARGUMENTS: 
            seq: list of items
            size: 
        process list in batches of size 
        eg. for group in chunker(list, 100):
                    print(group)
    """
    return (seq[pos:pos + size] for pos in range(0, len(seq), size))

def multithreading_list(func, input_list, *args, max_workers=cores, **kwargs):
    """
        ARGUMENTS
            func: function we want to execute in multithreading mode on a list of items
            input_list: list of items. Items from this list are passed as the first positional arg of func
            *args: additional positional args we want to pass to func
            max_workers: amount of threads used. Defaults to nbr of physical cores of the machine
            **kwargs: additional kwargs we want to pass to func

        INFO
            Use for heavy IO operations
            https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor

        RETURNS
            (dict(item: func return value ), dict(item: func error ))

            ! item is the item from the list we pass as first positional argument of the function. 
    """
    results = {}
    errors= {}
    with ThreadPoolExecutor(max_workers) as executor:
        future_to_x = {executor.submit(func, x, *args, **kwargs): x for x in input_list}
        for future in as_completed(future_to_x):
            x = future_to_x[future]
            # If an exception was raised while executing the task, it will be re-raised when calling the result() function or can be accessed via the exception() function.
            try:
                result = future.result()
                results[x] = result
            except Exception as exc:
                errors[x] = exc
                print('%r generated an exception: %s' % (x, exc))
    return results, errors

def multiprocessing(func, input_list, *args, max_workers=cores, **kwargs):
    """
        ARGUMENTS
            func: function we want to execute in multiprocessing mode on a list of items
            input_iterable: 
                    list of items. Each item can be another list. Items are passed as the first positional arg of the func.
                    In case of a list of lists, the inner list is unpacked an passed as multiple positional args. 
                    Only use list of list in case you want to pass several specific args. In case you want to pass common args, use *args
            *args: additional common positional args we want to pass to func
            max_workers: amount of threads used. Defaults to nbr of physical cores of the machine
            **kwargs: additional kwargs we want to pass to func

        INFO
            Use for heavy CPU operations
            https://docs.python.org/3/library/concurrent.futures.html#processpoolexecutor

        RETURNS
            (dict(item: function return value ), dict(item: function error ))

            !!! item is the item from the list we pass as first positional argument of the function. 

        ! multiprocessing list of list not tested yet
    """
    results = {}
    errors= {}
    with ProcessPoolExecutor(max_workers) as executor:
        if isinstance(input_list[0], list):
            future_to_x = {executor.submit(func, *x, *args, **kwargs): x for x in input_list}
        else:
            future_to_x = {executor.submit(func, x, *args, **kwargs): x for x in input_list}
        for future in as_completed(future_to_x):
            x = future_to_x[future]
            try:
                result = future.result()
                results[x] = result
            except Exception as exc:
                errors[x] = exc
                print('%r generated an exception: %s' % (x, exc))
    return results, errors

def valid_date(x: str) -> bool:
        """ 
            Checks whether a string has date format %Y-%m-%d (ie. 2020-01-01). 
            
            Returns datetime object or raises a ValueError
        """
        try:
            return datetime.strptime(x, "%Y-%m-%d")
        except ValueError as e:
            msg = "Not a valid date: {0!r}".format(x)
            raise argparse.ArgumentTypeError(msg) from e

def parse_args() -> Dict:
    """
        Parses command line arguments and returns dictionary of arguments which can be accessed via dot notation

        ARGUMENTS
            '--env', '-e', default ='local', type = str, choices = ['local', 'dev', 'prd']
            '--type', '-t', default = 'incremental', type = str, choices = ['event', 'incremental', 'full']
            '--airflow_execution_date', '-d', default = '2020-01-01', type = valid_date
            '--task', '-task', type = str, required=True
            '--event_bucket', type=str
            '--event_prefix', type=stR

    """
    # create parser
    parser = argparse.ArgumentParser()

    # add arguments
    parser.add_argument('--env', '-e', default ='local', type = str, choices = ['local', 'dev', 'prd'])
    parser.add_argument('--type', '-t', default = 'incremental', type = str, choices = ['event', 'incremental', 'full_refresh'])
    parser.add_argument('--airflow_execution_date', '-d', default = '2020-01-01', type = valid_date)
    parser.add_argument('--task', '-task', type = str, required=True)
    parser.add_argument('--event_bucket', type=str )
    parser.add_argument('--event_prefix', type=str)

    # parse arguments
    args = parser.parse_args()

    # manual error handling
    if args.type == 'event' and ( args.event_bucket is None or args.event_prefix is None ):
        parser.error("In case pipeline_type = 'event', both event_bucket and event_prefix need to be included")

    return args


import logging
from functools import wraps
import time
import sys

def retry(retries: int, **fkwargs):
    """
        Retry decorator with adjustable retries. 

        USAGE:

        @retry(retries = x, error = "Failure executing Message")
        def function(..., **kwargs)
            ...
            return

        ! Ensure you setup **kwargs in the function you're wrapping. The attempt number is passed to this function as a kwarg
    
    """
    def _decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for attempt in range(retries):
                try:
                    return func(*args, attempt = attempt, **kwargs)
                except Exception as e:
                    logging.warning(f"Error: {fkwargs['error']}, Attempt: {attempt}, Error: {e}")
                    if attempt < retries-1:
                        time.sleep((attempt+1)*(attempt+1))
            logging.error(f"Failed after {retries} attempts. Exiting...")
            sys.exit(1)
        return wrapper
    return _decorator


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pyveb']

package_data = \
{'': ['*']}

install_requires = \
['attrdict>=2.0.1,<3.0.0',
 'boto3==1.24.24',
 'numpy==1.21.2',
 'pandas==1.3.2',
 'psutil>=5.9.0,<6.0.0',
 'psycopg2-binary==2.9.1',
 'pyarrow==5.0.0',
 'pyodbc==4.0.30',
 'pyspark>=3.0.0,<4.0.0',
 'pyyaml==6.0',
 'requests==2.27.1',
 's3fs==0.4.2',
 'selenium>=4.5.0,<5.0.0',
 'simple-ddl-parser==0.26.5',
 'webdriver-manager==3.4.1',
 'xlrd>=2.0.0,<3.0.0']

setup_kwargs = {
    'name': 'pyveb',
    'version': '0.2.16',
    'description': 'Package containing common code and reusable components for pipelines and dags',
    'long_description': '# General \n\nPackage containing resuable code components for data pipelines and dags deployed to pypi.\n\n# Usage\n\nInstall/Upgrade locally: \n\n```\n$ pip3 install pyveb\n$ pip3 install pyveb --upgrade\n```\n\nImport in python\n\n```\nimport pyveb\nfrom pyveb import selenium_client\n```\n\n# Update package\n\nPackage is automaticly deployed to pypi via github actions. Just commit and open a pull request. During the action workflow, the version will be automatically bumped and updated pyproject.toml is commited back. \n\n! in case a dependency is added to pyproject.toml, no workflow is started unless there are also changes to src/pyveb/** \n\n\n\n\n\n\n\n\n\n',
    'author': 'pieter',
    'author_email': 'pieter.de.petter@veb.be',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)




import ftplib
import sys
import logging
# import s3_client
import os

class ftpClient():

    def __init__(self, ftp_folder:str, env:str, customer:str, s3_client:any, **kwargs) -> None: 
        """
            ARGUMENTS:
                ftp_folder: in case we need to drop files in a subfolder, ie. url/folder/
                env: local, dev or prd
                bucket: bucket containing files we want to upload to ftp
                customer: esight or eportal

            RETURNS:
                None

            ADDITIONAL INFO:
                currently only tested for esight/eportal flow but should work in general
        """
        if env == 'local':
            try:
                self.url = os.environ[f'{customer.upper()}_URL']
                self.user = os.environ[f'{customer.upper()}_USER']
                self.password = os.environ[f'{customer.upper()}_PASSWORD']
            except Exception as e: 
                logging.error(f'Issue fetchig ftp credentials from local env: {e}')
                sys.exit(1)
        else: 
            try:
                self.url=os.environ['FTP_URL']
                self.user = os.environ['FTP_USER']
                self.password = os.environ['FTP_PASSWORD']
            except Exception as e:
                logging.error(f'Issue fetchig ftp credentials: {e}')
                sys.exit(1)
        self.ftp_folder = ftp_folder
        self.env = env
        self.timeout = kwargs.get('timeout',120)
        self.ftp_session = self._create_ftp_session()
        self.s3client = s3_client
        return None

    def _create_ftp_session(self):
        try:
            ftp_session = ftplib.FTP(self.url, timeout=self.timeout)
            ftp_session.login(self.user, self.password)
            if self.ftp_folder:
                ftp_session.cwd(self.ftp_folder)
        except Exception as e:
            logging.error(f'Issue establishing ftp connection for {self.url}: {e}')
            sys.exit(1)
        return ftp_session

    def close_ftp_session(self):
        try: 
            self.ftp_session.quit()
        except Exception as e:
            logging.error('issue executing ftp_session.quit(), executing hard close ftp_session.close()')
            # https://pd.codechef.com/docs/py/2.7.9/library/ftplib.html
            self.ftp_session.close()

    def upload_to_ftp(self, file):
        memory_csv = self.s3client.download_s3_to_memory(file)
        if self.env == 'prd':
            file_name = 'VEB_'+file.split('/')[-1]
        else:
            file_name = 'VEB_TEST_'+file.split('/')[-1]
        self.ftp_session.storbinary(f'STOR {file_name}', memory_csv)



import boto3
from urllib.parse import unquote
import os, sys, time
import logging
from io import BytesIO
import pandas as pd

class s3Client():

    def __init__(self, s3_bucket:str) -> None:
        self.bucket_name = s3_bucket
        self.bucket = self._create_bucket_resource()
        self.client = self._create_client()

    def _create_client(self):
        logging.info('Creating boto3 s3 client')
        s3_client = boto3.client("s3")
        logging.info('Succesfully created s3 client')
        return s3_client

    def _create_bucket_resource(self):
        logging.info('Creating boto3 s3 bucket resource')
        s3 = boto3.resource('s3')
        bucket = s3.Bucket(self.bucket_name)
        logging.info('Succesfully created bucket resource')
        return bucket

    def list_files(self, s3_prefix:str, file_type:str, max_files=1000000000000, list_empty_files=True) -> list:
        """
            ARGUMENTS
                s3_prefix: folder/subfolder/
                file_type: eg. csv, xlsx, parquet
                max_files: optional, unless max_files is specified, all files are returned. 
                list_empty_files: if False, files with size 0 are not listed

            RETURNS 
                List of files from s3_prefix, filtering by type. Files are fetched per chunks of 1000. 
            
            ADDITIONAL INFO 
            Files are automatically unquoted and returned as full path ( ie. s3:://bucket/s3_prefix/filename.extension)
            
            List_empty_files kwarg is set to True by default to enable backwards compatibility. 
        """
        kwargs = {
            "Bucket": self.bucket_name,
            "Delimiter": "|",
            "EncodingType": "url",
            "MaxKeys": 1000,
            "Prefix": s3_prefix,
            "FetchOwner": False
        }
        truncated = True
        files = []
        counter = 0
        while truncated and counter < max_files:
            counter += 1000
            response = self.client.list_objects_v2(**kwargs)
            truncated = response["IsTruncated"]
            if "Contents" in response:
                if list_empty_files:
                    files.extend(record["Key"] for record in response["Contents"])
                else:
                    files.extend(record["Key"] for record in response["Contents"] if record["Size"] > 0)
                if truncated:
                    kwargs["ContinuationToken"] = response["NextContinuationToken"]
        files = files[:max_files]
        files = [x for x in files if file_type in x]
        files = [unquote(x) for x in files]
        files = ['s3://'+self.bucket_name+'/'+x for x in files]
        return files

    def list_files_bigger_than(self, s3_prefix:str, file_size_bytes:int = 10000000, file_type:str = 'parquet') -> list:
        """
            ARGUMENTS
                s3_prefix: folder/subfolder/
                file_size_bytes: optional, eg. 10000000 (note: 10 000 000 = 10mb)
                file_type: optional, eg. csv, xlsx, parquet
            RETURNS 
                List of files from filtered by type and strictly bigger than file_size_bytes
        """
        big_files = []  
        for file in self.bucket.objects.filter(Prefix=s3_prefix):
            if file.size > file_size_bytes and f'.{file_type}' in file.key:
                file_name = f's3://{self.bucket_name}/{file.key}'  
                big_files.append(file_name)
        return big_files

    def _delete_prefix(self, s3_prefix:str ) -> None:
        """
            Delete all objects with a certain prefix
        """
        logging.info(f'Starting delete operation for prefix {s3_prefix}')
        self.bucket.objects.filter(Prefix=s3_prefix).delete()
        logging.info(f'Succesfully delete all objects for prefix {s3_prefix}')
        return
    
    def _prefix_exist_and_not_empty(self, s3_prefix:str) -> bool:
        '''
            S3_prefix should exist. 
            S3_prefix should not be empty.

            Returns boolean
        '''
        if not s3_prefix.endswith('/'):
            s3_prefix += '/'
        resp = self.client.list_objects(Bucket=self.bucket_name, Prefix=s3_prefix, Delimiter='/', MaxKeys=1)
        return 'Contents' in resp, 'CommonPrefixes' in resp

    def delete_prefix_if_exist(self, s3_prefix:str) -> None:
        """
            ARGUMENTS
                s3_prefix: folder/subfolder/

            RETURNS 
                None

            ADDITIONAL INFO
                Checks whether the prefix contains at least 1 file or folder. If not, prefix is considered non-existent.
                If yes, prefix ( ie all folders, subfolders and files) will be deleted. 
        """
        try: 
            data_exist, subdir_exist = self._prefix_exist_and_not_empty(s3_prefix)
            if data_exist or subdir_exist:
                self._delete_prefix(s3_prefix)
                logging.warning(f'Found and deleted prefix {s3_prefix}.')
            else:
                logging.warning(f'Partition {s3_prefix} does not exist and will be created')
        except Exception as e:
            logging.error('Issue checking if prefixs exists already and/or deleting. Exiting...')
            logging.error(e)
            sys.exit(1)
        return

    def upload_local_file(self, local_file:str, s3_prefix: str) -> None:
        """
            ARGUMENTS
                local_file: ./data/file.ext
                s3_prefix: target_folder/target_subfolder

            RETURNS
                None

            ADDITIONAL INFO
                Use this function to upload a single file from local to s3. The local filename will be retained. 
        """
        target_file_name = local_file.split('/')[-1]
        target_file = f'{s3_prefix}{target_file_name}'
        try: 
            logging.info('Uploading file ...')
            self.client.upload_file(local_file, self.bucket_name, target_file)
            logging.info(f'Uploaded {local_file} to {target_file}')
        except Exception as e: 
            logging.error(f'Issue uploading {local_file} to {target_file}')
            logging.error(e)
            sys.exit(1)
        return

    def df_to_parquet(self, df: pd.DataFrame, s3_prefix:str, s3_file_name:str, add_timestamp:bool = True ) -> None:
        """
            ARGUMENTS
                df: pandas dataframe
                s3_prefix: target_folder/target_subfolder
                s3_file_name: name of the file without extension.  
                add_timestamp: if True, s3_file_name will be prefxied with a timestamp

            RETURNS
                None

            ADDITIONAL INFO
                DF will be read into memory and stored in S3 as parquet under key s3_prefix/1562388.0020_s3_file_name.parquet
        """
        logging.info('Starting upload')
        parquet_buffer = BytesIO()
        df.to_parquet(parquet_buffer, index=False, allow_truncated_timestamps=True)
        if add_timestamp:
            timestamp = round(time.time(), 4)
            s3_key = f'{s3_prefix}{timestamp}_{s3_file_name}.parquet'
        else:
            s3_key = f'{s3_prefix}{s3_file_name}.parquet'
        parquet_buffer.seek(0)
        self.bucket.upload_fileobj(parquet_buffer, s3_key)
        logging.info(f'Uploaded dataframe to {s3_key}')
        return

    def df_to_csv_s3(self, df: pd.DataFrame, s3_prefix:str, s3_file_name:str, delimiter=",") -> None:
        """
            ARGUMENTS
                df: pandas dataframe
                s3_prefix: target_folder/sub_folder
                s3_file_name: name of the file without extension. A timestamp will be added within the function. Eg, file_name = pipeline_name
                delimiter: df will be converted to a csv file with the provided delimiter
            
            RETURNS
                None

            ADDITIONAL INFO
                Dataframe will be read into memory and stored in S3 as parquet under key s3_prefix/1562388.0020_s3_file_name.csv
        """
        timestamp = round(time.time(), 4)
        csv_buffer = BytesIO()
        df.to_csv(csv_buffer, index=False, sep=delimiter)
        s3_key = f'{s3_prefix}{timestamp}_{s3_file_name}.csv'
        csv_buffer.seek(0)
        self.bucket.upload_fileobj(csv_buffer, s3_key)
        return 

    def download_s3_to_local_file(self, file:str, file_extension = 'parquet', download_path='./data') -> str:
        if not download_path.endswith('/'):
            download_path = f'{download_path}/'
        timestamp = round(time.time(), 4)
        local_file = f'{download_path}/{timestamp}.{file_extension}'
        if not os.path.exists(download_path):
            os.mkdir(download_path)
        file = file.split('//')[1].split('/',1)[1]
        try:
            self.bucket.download_file(file, local_file)
        except Exception as e:
            logging.error(f'Issue downloading {file}. Exiting...')
            logging.error(e)
            sys.exit(1)
        return local_file

    def download_s3_to_memory(self, file:str):
        """
            !! only tested for csv files 

            Returns: 
                filestream object with pointer at position 0
        """
        try: 
            file = file.split('//')[1].split('/',1)[1]
            obj = self.bucket.Object(file)
            file_stream = BytesIO()
            obj.download_fileobj(file_stream)
            # https://stackoverflow.com/questions/61690731/pandas-read-csv-from-bytesio
            file_stream.seek(0)
        except Exception as e:
            logging.error(f'Issue downloading {file} to memory stream. Exiting...')
            logging.error(e)
            sys.exit(1)
        return file_stream



class externalS3Client():

    def __init__(self, aws_access_key_id_name:str, aws_secret_access_key_name:str , ext_bucket:str):
        """ 
            ARGS
                aws_access_key_id_name: environment variable holding aws_access_key_id
                aws_secret_access_key_name: environment variable holding aws_secret_access_key_name
                ext_bucket: external bucket name without 's3://" eg, ext_bucket_name
        
        """
        self.bucket = ext_bucket
        try:
            aws_access_key_id=os.environ[aws_access_key_id_name]
            aws_secret_access_key = os.environ[aws_secret_access_key_name]
        except Exception: 
            logging.error('AWS credentials for external account not found. Exiting...')
            sys.exit(1)

        session_ext = boto3.Session(
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key
        )
        self.s3 = session_ext.resource('s3')

    def put_object(self, s3_prefix:str, local_file:str, ext_file_name:str):
        """ 
            puts an object inside the external bucket
            inside the specified directory, with the specified filename
        """
        data = open(local_file, 'rb')
        try:
            self.s3.Bucket(self.bucket).put_object(Key=f"{s3_prefix}/{ext_file_name}", Body=data)
            logging.info(f"stored in external s3 bucket: {self.bucket}")
        except Exception as e:
            logging.error(f"Could not store file: {ext_file_name} in external s3 bucket")
            logging.error(e)
            sys.exit(1)

